#include "eopt.hh"
#include "verbose.hh"

#define max(a,b) (((a)>(b))?(a):(b))

EOptimist::EOptimist(Architecture *_architecture, float _delta_e)
  : architecture(_architecture), delta_e(_delta_e),
    cut(0), improved(0), merged(0), n_enodes(0)
{
}

EOptimist::~EOptimist()
{
}

Schedule *EOptimist::schedule(DAG & _dag, Flags & _flags)
{
  ZSet zset;
  Matcher matcher(architecture);
  cut = improved = merged =  n_enodes = 0;

  NodeAttr *init_node_attr = get_init_node_attr(zset, _dag);
  TimeProfile *init_time_profile = new TimeProfile(architecture->get_reserv_tab_size());
  ENode *init_enode = 
    new ENode(init_node_attr, init_time_profile, architecture);

  int num_vertices = boost::num_vertices(_dag);
  //Reduce ENode: compute ADDR and CNST nodes to their respective residences
  int num_reduced = reduce_init_enode(*init_enode, matcher, zset, _dag);
  num_vertices -= num_reduced;

  coord_t coord(GRID_DIM), max_coord(GRID_DIM);

  solution_space_t solution_space;
  coord[L_AXIS] = 0;
  coord[E_AXIS] = 0;
  max_coord = coord;
  solution_space.insert(init_enode, coord);
  VERB1 << "num_vertices (- num_reduced): " << num_vertices << "\n";

  for(unsigned int e = 0; e <= max_coord[E_AXIS]; e++) {

    for(unsigned int l = 0; l < num_vertices; l++) {

      coord[E_AXIS] = e;
      coord[L_AXIS] = l;
      //      std::set< ENode * >::iterator gi, gi_end;
      grid_elem_t::iterator gi, gi_end;
      for(tie(gi, gi_end) = solution_space.elements(coord);
	  gi != gi_end; ++gi) {
	ENode *current_enode =  *gi;

	vertex_set_t::iterator zi, zi_end;
	NodeAttr *current_node_attr = current_enode->get_node_attr();
	zi = current_node_attr->z.begin();
	zi_end = current_node_attr->z.end();
	for( ; zi != zi_end; ++zi) {
	  vertex_id u_id = *zi;
	  Matcher::match_map_iter mi, mi_end;
	  //For all match y covering (u_id, current_enode, & chi, & instr)
	  for(tie(mi, mi_end) = matcher.matches(u_id, *current_enode, _dag);
	      mi != mi_end; ++mi) {

	    std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;

	    for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
		iso_iter != p.second.end(); ++iso_iter) {
	      vertex_set_t chi;
	      chi.clear();
	      for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
		  vi != (iso_iter->iso_map).end(); ++vi) {
		chi.insert(vi->second);
	      }
	      update_space(current_enode, mi->first, chi, coord, solution_space, max_coord,
			   iso_iter->root, iso_iter->arg_map, *(p.first), zset, _dag);
	    }
	  }
	}

	//Transfers
	SpaceProfile::iterator spi = current_enode->get_space_profile()->begin(),
	  spi_end = current_enode->get_space_profile()->end();
	for( ; spi != spi_end; ++spi) {
	  vertex_id u = spi->first;
	  std::set< int > & rcs = spi->second;
	  for(std::set< int >::iterator src_iter = rcs.begin(); src_iter != rcs.end();
	      ++src_iter) {
	    std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > > dests =
	      architecture->get_transfer_destinations(*src_iter);
	    if(!dests.first) {
	      continue;
	    }
	    std::vector< std::pair< int, TargetInstruction * > >::iterator dti_iter = 
	      dests.second.begin(), dti_iter_end = dests.second.end();
	    for( ; dti_iter != dti_iter_end; ++dti_iter) {
	      if(rcs.find(dti_iter->first) == rcs.end()) {
		//node u is not in residence class of the destination of the transfer instruction
		//add transfer instruction
		node_map_t dummy_node_map;
		pattern_t dummy_pattern;
		update_space(current_enode, dti_iter->second, vertex_set_t(), coord, 
			     solution_space, max_coord, u, dummy_node_map, dummy_pattern, zset, _dag);
	      }
	    }
	  }
	}
      }
    }
  }

  //Solution
  std::vector< ENode * > candidates;
  coord_t stop_coord(GRID_DIM);
  stop_coord[L_AXIS] = num_vertices;
  //  std::set< ENode * >::iterator gi, gi_end, min_gi, gi_iter;
  grid_elem_t::iterator  gi, gi_end, min_gi, gi_iter;
  for(int e = 0; e <= max_coord[E_AXIS]; e++) {
    stop_coord[E_AXIS] = e;
    tie(gi, gi_end) = solution_space.elements(stop_coord);
    if(gi != gi_end) {
      break;
    }
  }


  assert(gi != gi_end); //Found solutions
  min_gi = gi;

  for(gi_iter = gi; gi_iter != gi_end; ++gi_iter) {
    if((*min_gi)->get_e() > (*gi_iter)->get_e()) {
      min_gi = gi_iter;
    }
  }
  VERB1 << "tau = " << (*min_gi)->get_tau()
	<< "\nn_enodes: " << n_enodes
	<< " merged: " << merged
	<< " cut: " << cut
	<< " improved: " << improved
	<< "\nenergy: " << (*min_gi)->get_e() << "\n";

  delete init_enode;
  delete init_time_profile;

  Schedule *s = (*min_gi)->get_schedule();
  return s;
}

NodeAttr *EOptimist::get_init_node_attr(ZSet & _zset, DAG & _dag)
{
  graph_traits< DAG >::edge_iterator ei, ei_end;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  indeg_t indeg;
  vertex_set_t z;
	
  //All vertices start with indegree 0
  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    indeg[*vi] = 0;
  }
  //Initial indegrees
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id v_id = target(*ei, _dag);
    indeg[v_id]++;
  }	
  //Initial z
  indeg_t::iterator ii;
  for(ii = indeg.begin(); ii != indeg.end(); ++ii) {
    if(indeg[ii->first] == 0) {
      z.insert(ii->first);
    }
  }
	
  vertex_set_t alive;
	
  //Initial set of alive is empty if we only consider basic block level
  alive.clear();
  NodeAttr *node_attr = _zset.insert(z, indeg, alive);
  return node_attr;
}

int EOptimist::reduce_init_enode(ENode &_enode, Matcher _matcher, ZSet & _zset, DAG & _dag)
{
  NodeAttr *_node_attr = _enode.get_node_attr();
  vertex_set_t::iterator zi, zi_end;
  zi = _node_attr->z.begin();
  zi_end = _node_attr->z.end();
  int num_reduced = 0;

  for( ; zi != zi_end; ++zi) {
    vertex_id u_id = *zi;
    lcc::Node u = get_lcc_node(u_id, _dag);
    //For all nodes
    bool b_reduced = false;
    switch(generic(u->op)) {
    case lcc::ADDRF: case lcc::ADDRG: case lcc::ADDRL: case lcc::CNST:
      {
	num_reduced++;
	Matcher::match_map_iter mi, mi_end;
	//For all match of CNST, or ADDR
	for(tie(mi, mi_end) = _matcher.matches(u_id, _enode, _dag);
	    mi != mi_end; ++mi) {
	  std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;
	  
	  for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
	      iso_iter != p.second.end(); ++iso_iter) {
	    vertex_set_t chi;
	    chi.clear();
	    for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
		vi != (iso_iter->iso_map).end(); ++vi) {
	      chi.insert(vi->second);
	    }
	    assert(chi.size() == 1);

	    //FIXME see snode.cc
	    int height = 0;
	    const reserv_tab_t & reserv_tab = mi->first->get_reserv_tab();
	    for(int i = 0; i < reserv_tab.size(); i++) {
	      height = max(reserv_tab[i].size(), height); 
	    }
	    if(height > 0) {
	      std::cerr << "Reduce node with a reservation table!\n";
	    }
	    //update space profile only.
	    SpaceProfile *space_profile = _enode.get_space_profile();
	    space_profile->insert(_node_attr->alive, u_id, _dag, height,
				  mi->first->get_residence_class("dest1"), architecture);
	    if(!b_reduced) {
	      _node_attr = select(chi, *_node_attr, _zset, _dag);
	      b_reduced = true;
	    }
	  }
	}
      }
    }
  }
  _enode.set_node_attr(_node_attr);
  return num_reduced;
}

//Common method (from snode.cc)
NodeAttr *EOptimist::select(const vertex_set_t & _chi, 
			    const NodeAttr & _node_attr,
			    ZSet & _zset, DAG & _dag)
{
  vertex_set_t z(_node_attr.z);
  indeg_t indeg(_node_attr.indeg);
  vertex_set_t alive;

  vertex_set_t::const_iterator chi_iter = _chi.begin();
  for(; chi_iter != _chi.end(); ++chi_iter) {
    vertex_id u_id = *chi_iter;
    //successors of v_id are decremented
    successor_iterator pi, pi_end;
    for(tie(pi, pi_end) = successors(u_id, _dag);
	pi != pi_end; ++pi) {
      vertex_id v_id = *pi;
      indeg[v_id]--;
      if(indeg[v_id] == 0) {
	z.insert(v_id);
      }
    }
  }
  //Remove chi nodes from z
  for(chi_iter = _chi.begin(); chi_iter != _chi.end(); ++chi_iter) {
    assert(indeg[*chi_iter] == 0);
    z.erase(*chi_iter);
  }
  //Alive set
  graph_traits< DAG >::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id u_id = source(*ei, _dag),
      v_id = target(*ei, _dag);
    if((indeg[u_id] == 0) && (z.find(u_id) == z.end())) {
      if((indeg[v_id] != 0) || (z.find(v_id) != z.end())) {
	//Target in zero-indegree set
	alive.insert(u_id);
      }
    }			
  }
	
  NodeAttr *node_attr = _zset.insert(z, indeg, alive);
  return node_attr;
}

void EOptimist::update_space(const ENode *_current_enode,
			     TargetInstruction *_target_instruction,
			     const vertex_set_t & _chi,
			     const coord_t & _current_coord,
			     solution_space_t & _solution_space,
			     coord_t & _max_coord,
			     vertex_id & _root,
			     node_map_t & _arg_map,
			     pattern_t & _pattern,
			     ZSet & _zset,
			     DAG & _dag)
{
  NodeAttr *succ_node_attr = 
    select(_chi, *_current_enode->get_node_attr(), _zset, _dag);
  ENode *succ_enode = NULL;

  std::vector< ENode * > candidates;

  int tau = _current_enode->get_tau();
  int offset = 0; //insert between t and tau
  int succ_tau = 0;
  std::pair< bool, float > p;

  int height = 0; //FIXME: part of the pattern
  const reserv_tab_t & reserv_tab = _target_instruction->get_reserv_tab();
  for(int i = 0; i < reserv_tab.size(); i++) {
    height = max(reserv_tab[i].size(), height); 
  }

  do {
    //Loop over all possible delays
    succ_enode = new ENode(*_current_enode);
    succ_enode->set_node_attr(succ_node_attr);

    p = succ_enode->insert(_chi, _root, _dag, _arg_map, _pattern, 
			   _target_instruction, architecture, offset);
    offset++;
    if(!p.first) {
      delete succ_enode;
      cut++;
      break;
    }
    succ_tau = succ_enode->get_tau();
    if((succ_tau > tau) && (candidates.size() > 1)) {
      delete succ_enode;
    } else {
      candidates.push_back(succ_enode);
    }
  } while(p.first && (tau >= succ_tau) && (height > 0));

  coord_t check_coord(GRID_DIM), dest_coord(GRID_DIM);
  check_coord = _current_coord;
  int dest_l = _current_coord[L_AXIS] + _chi.size();
  dest_coord[L_AXIS] = dest_l;
  int nb_candidates = candidates.size();
  for(int i = 0; i < nb_candidates; i++) { 
    succ_enode = candidates[i];
    float energy = succ_enode->get_e();

    dest_coord[E_AXIS] = (int)(energy / delta_e);
    check_coord[L_AXIS] = dest_l;


    check_coord[E_AXIS] = _current_coord[E_AXIS];
    ENode *enode_prime = NULL;
    unsigned int max_energy = _max_coord[E_AXIS];

    //    std::pair< std::set< ENode *>::iterator, bool > ep;
    std::pair< grid_elem_t::iterator, bool > ep;

    for( ; check_coord[E_AXIS] <= max_energy; check_coord[E_AXIS]++) {
      ep = _solution_space.exist(succ_enode, check_coord);
      if(ep.second) {
	break;
      }
    }

    if(ep.second) {
      ENode *tmp_enode = *ep.first;
      assert(*tmp_enode == *succ_enode);
      if(succ_enode->get_e() < tmp_enode->get_e()) {
	//Improved intermediate solution
	_solution_space.erase(*(ep.first), check_coord);
	_solution_space.insert(succ_enode, dest_coord);
	improved++;
      } else {
	//New solution is equivalent or worse (forget it)
	delete succ_enode;
	merged++;
      }
    } else {
      //New intermediate solution
      ep = _solution_space.insert(succ_enode, dest_coord);
      assert(ep.second); //Insertion succeeded
      n_enodes++;
      //Update max_coord
      _max_coord[E_AXIS] = max(_max_coord[E_AXIS], dest_coord[E_AXIS]);
    }
  }
  candidates.clear();
}
