#ifndef __GAPOPULATION_HH__
#define __GAPOPULATION_HH__

//#include <sys/int_types.h>
#ifdef _WPT_
#include <pthread.h>
#include <semaphore.h>
#endif

#include <csignal>
#include <vector>
#include <deque>
#include "pudag.hh"
#include "flags.hh"
#include "snode.hh"
#include "matcher.hh"
#include "zset.hh"

#include "gautility.hh"
#include "gafitness.hh"
#include "gaindividual.hh"
#include "gainstcand.hh"
#include "gagnode.hh"

namespace GA
{

void * thread_create_schedule( void * arg );

typedef std::deque<Individual*> individuals_t;
typedef std::vector<InstructionCandidate> instruction_candidates_t;


/**
 * A class for representing a population of individuals and their
 * posiblity to evolve.
 */
class Population
{
public:
  enum SurvivalSelectionMethod {
    FITNESS_PROPORTIONAL_SELECTION,
    TRUNCATION_SELECTION,
  };
  

public:
  Population(unsigned int _size, Architecture*, DAG&, Flags&);
  ~Population();
  
  void initialize();
  unsigned int get_size() const { return size; }
  Individual* get_individual(unsigned int i) { return individuals[i]; }
  Individual* get_best_individual();
  void evolve(unsigned int max_times);
  fitness_t advance();
  void inc_recreated_schedule_count() {++recreated_schedule_count;}
  /* mater: moved to public so I could parallelize (I need static
     pointers). */
  Schedule* create_schedule(Individual* ind, bool new_individual = false, bool simple_heuristics = false, bool parallel = false);
  Individual* generate_individual();

private:
  // TODO: Implement copy constructor and assignment operator, then make to public.
  Population(Population& p);
  Population& operator=(const Population& p);
  
  std::pair<Individual*, Individual*> crossover(Individual* parent1, Individual* parent2);
  void mutate(individuals_t& _individuals);
  void mutate_node(Individual* ind);
  void mutate_inst(Individual* ind);
  void mutate_transfer(Individual* ind);
  
  void delete_individuals(individuals_t& _individuals);
  void calculate_fitness(individuals_t& _individuals);
  void create_schedules(individuals_t& _individuals);
  
  Individual* select_parent(individuals_t& _individuals);
  Individual* binary_tournament_selection(individuals_t& _individuals);
  
  void select_survivors(individuals_t& individuals, unsigned int num);
  void roulette_wheel_selection(individuals_t& individuals, unsigned int num);
  void truncation_selection(individuals_t& individuals, unsigned int num);
  
  void build_instruction_priorities(Individual* ind);
  void build_zorder(Individual* ind);
  void build_transfer_priorities(Individual* ind);

  instruction_candidates_t::iterator find_transfer_candidate(instruction_candidates_t& candidates,
                                                             Individual* ind, vertex_id last_selected,
                                                             unsigned int num_previous_transfers);
  instruction_candidates_t::iterator find_instruction_candidate(instruction_candidates_t& candidates,
                                                                Individual* ind, vertex_id u_id);
  
  NodeAttr* select(const vertex_set_t& _chi, const NodeAttr& _node_attr, DAG& _dag, ZSet * zset);
  NodeAttr* get_init_node_attr(DAG& _dag, ZSet * zset);
  void match_node(instruction_candidates_t& candidates, vertex_id u_id, SNode* current_snode,
                  DAG& _dag, std::list<SNode*>& snodes, GNode* _gnode, Individual* _ind);
  void transfers(instruction_candidates_t& candidates, SNode * current_snode,
                 std::list<SNode*>& snodes, GNode* _gnode, Individual* _ind);
  bool update_space(SNode* current_snode, InstructionCandidate candidate,
                    std::list<SNode*>& snodes, GNode* gnode, ZSet * zset);
  InstructionType select_instruction(instruction_candidates_t& candidates, vertex_id u_id,
                                     vertex_id last_selected, SNode* current_snode,
                                     std::list<SNode*>& snodes, GNode* gnode,
                                     Individual* ind, unsigned int num_previous_transfers,
				     ZSet * zset);
  zorder_t::iterator get_first_z_element_in_zorder(const vertex_set_t& z, zorder_t& zorder);
  
  
  individuals_t individuals;
  
  int generation;
  Individual* best_individual;
  
  enum SurvivalSelectionMethod survival_method;
  
  int STABLE_GENERATIONS_LIMIT;
  double MUTATE_NODE_CHANCE;
  double MUTATE_INST_CHANCE;
  double MUTATE_TRANSFER_CHANCE;
  bool PARENTS_SURVIVE;
  bool MUTATE_ALLOW_MULTIPLE;
  double NEW_CHILDREN_PERCENTAGE;
  
  bool RANDOM_SEARCH;
  
  bool BENCH; // Only prints the result numbers.
  
  unsigned int size;
  Architecture *architecture;
  DAG& dag;
  Flags& flags;
  Matcher *matcher;
  
  int recreated_schedule_count;
  
  typedef bool (*condition)(std::vector< lcc::Node >);
  condition *conditions;
#ifdef _WPT_
  pthread_t *child;    //array of slaves
  struct thread_create_schedule_arg *arg; //array of slave args
#endif
};

#ifdef _WPT_
class thread_create_schedule_arg{
public:
  Individual ** ind;
  Population * pop;
  int nr;
  sem_t * start, * done;
};
#endif

};



#endif //__GAPOPULATION_HH__
