#ifndef _TP_HH_
#define _TP_HH_

/***
 * A time profile is a vector (size n, number of functional units) of 
 * bit vectors (vector< bool >). Time profiles store reservation tables of
 * matched instructions for a given set of DAG nodes. We do not store the
 * target instruction itself, but only the nodes.
 * Nodes under computation, i.e. still in the time profile are stored in the
 * nodes (set< vertex_id >). We associate a map for each node, reporting
 * latency before the result is available.
 * Two time profiles are equal if their respective bit vectors are equal.
 */
 
#include <iostream>
#include <vector>
#include <set>
#include <map>
#include "c.hh"
#include "pudag.hh"

const bool EMPTY = false;
const bool BUSY = true;
const int MP = 15;

/******************************************************************************
 * class TimeProfile                                                          *
 * Resources are indexed by [Resource number][Time]                           *
 ******************************************************************************/
class TimeProfile {

public:
  TimeProfile(int _size);
  TimeProfile(const TimeProfile &);
  ~TimeProfile();
  bool operator==(const TimeProfile &) const;
  //returns the number of time slots the time profile advanced
  int insert(const vertex_set_t & _nodes, const reserv_tab_t & _reserv_tab, DAG & _dag);
  //return the maximum pipeline size of the given profile: tau = ref_t + depth
  int get_depth(void);
  vertex_set_t & get_nodes(void) { return nodes; }
  friend std::ostream & operator<<(std::ostream &, const TimeProfile &);
  void advance(int);
  size_t get_hash() const { return hash_value; }

  // Oskar:
  std::map< vertex_id, int >& get_latency_map() { return latency_map; }
  // mater:
  reserv_tab_t & get_rtab() const { return *rtab; }
  int size() const {return rtab->size();}

private:
  reserv_tab_t * rtab; //mater
  vertex_set_t nodes;
  /* Stores relative times when results are ready. This one is confusing. */
  std::map< vertex_id, int > latency_map;
  /* Stores absolute times when results are ready */
  std::map< vertex_id, int > ready_map;
  int get_ready(vertex_id v) const;

  void insert_reserv_tab(const vertex_set_t & _nodes,
			 const reserv_tab_t & _reserv_tab, int _latency);
  void compute_hash(void);

  static const size_t primes[MP][MP]; //For hashing, matrix of prime numbers
  size_t hash_value;
  int current_time;
};

/*
//std::ostream & operator<<(std::ostream &, const Slot &);
//std::ostream & operator<<(std::ostream &, const UnitTimeProfile &);
*/
std::ostream & operator<<(std::ostream &, const TimeProfile &);

#endif //_TP_HH_
